import Metal
import MetalKit
import simd

enum RendererError: Error {
    case badVertexDescriptor
}

class Renderer: NSObject, MTKViewDelegate {

    var _mtkView: MTKView
    let _device: MTLDevice
    private let _commandQueue: MTLCommandQueue
    private var _pipelineState: MTLRenderPipelineState
    private var _depthState: MTLDepthStencilState
	var _bgColor = MTLClearColor(red:0,green:0,blue:0,alpha:1)
    private let _inFlightSemaphore = DispatchSemaphore(value: _maxBuffersInFlight)
    public var _projectionMatrix: matrix_float4x4 = matrix_float4x4()
	var _eye:simd_float3 = simd_float3(0,0,-500)
	var _center:simd_float3 = simd_float3(0,0,0)
	var _up:simd_float3 = simd_float3(0,1,0)
	private var _sampler:MTLSamplerState?
	private var _models : [VexilModel] = []
    public var _viewportSize: CGSize?

    init?(metalKitView: MTKView) {
        _mtkView = metalKitView
        self._device = _mtkView.device!
        guard let queue = self._device.makeCommandQueue() else {
            fatalError("makeCommandQueue Error")
        }
        self._commandQueue = queue

		_mtkView.depthStencilPixelFormat = MTLPixelFormat.depth32Float_stencil8
        _mtkView.colorPixelFormat = MTLPixelFormat.bgra8Unorm_srgb
        _mtkView.sampleCount = 1
		
        do {
            _pipelineState = try Renderer.buildRenderPipelineWithDevice(device: _device,
                                                                       metalKitView: _mtkView)
        } catch {
            fatalError("Unable to compile render pipeline state.  Error info: \(error)")
        }
        
        let depthStateDesciptor = MTLDepthStencilDescriptor()
        depthStateDesciptor.depthCompareFunction = MTLCompareFunction.less
        depthStateDesciptor.isDepthWriteEnabled = true
        guard let state = _device.makeDepthStencilState(descriptor:depthStateDesciptor) else {
            fatalError("makeDepthStencilState Error")
        }
        _depthState = state

		let samplerDescriptor:MTLSamplerDescriptor = MTLSamplerDescriptor()
		samplerDescriptor.minFilter = MTLSamplerMinMagFilter.nearest
		samplerDescriptor.magFilter = MTLSamplerMinMagFilter.linear
		samplerDescriptor.sAddressMode = MTLSamplerAddressMode.repeat
		samplerDescriptor.tAddressMode = MTLSamplerAddressMode.repeat
		_sampler = _device.makeSamplerState(descriptor: samplerDescriptor)

		super.init()
    }
	
	func loadModel(model:VexilModel) -> VexilModel {
		_models.append(model)
		model.setColor(device: _device)
		model.setTexture(device: _device)
		model.setMatrixAnimation(device: _device)
		return model
	}

    class func buildRenderPipelineWithDevice(device: MTLDevice,
                                             metalKitView: MTKView) throws -> MTLRenderPipelineState {
        let library = device.makeDefaultLibrary()
        
        let vertexFunction = library?.makeFunction(name: "vertexShader")
        let fragmentFunction = library?.makeFunction(name: "fragmentShader")
        
        let pipelineDescriptor = MTLRenderPipelineDescriptor()
        pipelineDescriptor.label = "RenderPipeline"
        pipelineDescriptor.sampleCount = metalKitView.sampleCount
        pipelineDescriptor.vertexFunction = vertexFunction
        pipelineDescriptor.fragmentFunction = fragmentFunction
        pipelineDescriptor.vertexDescriptor = MTLVertexDescriptor()
        
        pipelineDescriptor.colorAttachments[0].pixelFormat = metalKitView.colorPixelFormat
        pipelineDescriptor.depthAttachmentPixelFormat = metalKitView.depthStencilPixelFormat
        pipelineDescriptor.stencilAttachmentPixelFormat = metalKitView.depthStencilPixelFormat
        
        return try device.makeRenderPipelineState(descriptor: pipelineDescriptor)
    }
	
    func draw(in view: MTKView) {
        _ = _inFlightSemaphore.wait(timeout: DispatchTime.distantFuture)
        
        if let commandBuffer = _commandQueue.makeCommandBuffer() {
            let semaphore = _inFlightSemaphore
            commandBuffer.addCompletedHandler { (_ commandBuffer)-> Swift.Void in
                semaphore.signal()
            }
            let renderPassDescriptor = view.currentRenderPassDescriptor

			let colorAttachment = renderPassDescriptor?.colorAttachments[0]
			colorAttachment?.clearColor = _bgColor

			if let renderPassDescriptor = renderPassDescriptor, let renderEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor) {
				let viewMatrix = Transforms.lookAt(eye: _eye, center: _center, up: _up)

                /// Final pass rendering code here
                renderEncoder.label = "Primary Render Encoder"
                renderEncoder.pushDebugGroup("Draw Box")
                renderEncoder.setCullMode(.back)
                renderEncoder.setFrontFacing(.counterClockwise)
                renderEncoder.setRenderPipelineState(_pipelineState)
                renderEncoder.setDepthStencilState(_depthState)
				renderEncoder.setFragmentSamplerState(_sampler, index: 0)

				for model in _models {
					model.updateDynamicBufferState()
					model.updateGameState(projectionMatrix:_projectionMatrix,viewMatrix: viewMatrix)
					model.encoder(renderEncoder: renderEncoder)
				}
                
                renderEncoder.popDebugGroup()
                renderEncoder.endEncoding()
                
                if let drawable = view.currentDrawable {
                    commandBuffer.present(drawable)
                }
            }
            
            commandBuffer.commit()
        }
    }
    
    func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
        let aspect = Float(size.width) / Float(size.height)
        let dest = _mtkView.bounds
        _viewportSize = dest.size
        _projectionMatrix = Transforms.perspective_fov(fovy: 65, aspectRatio:aspect, nearZ: 0.1, farZ: 10000.0)
    }
}
